#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		     Network Request Dispatcher  (UUCPSEND)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	20 Feb 86   WJB  First functioning version.
 */

#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include "queue.h"
#include "qreplys.h"
#include <sys/stat.h>

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

time_t	timenow;
struct	request r;
int	ourpid;
char	reqtmp[] = "/tmp/reqtmpXXXXXX";
char	missingfmt[] = "Missing %s argument";
char	*workdir = 0;
int	options = 0;
int	oldformat = 0;
FILE	*netfp;

char	*index();
char	*strncpy();
char	*newtmp();
char	*strdup();
int	doflush();
int	dorestart();
FILE	*popen();
struct passwd	*getpwuid();
struct request	*gethdr();

extern	char *Qdatadir;

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	command[LINESIZ];
	char	linebuf[LINESIZ];
	char	newline[LINESIZ];
	char	datafile[PATHSIZ];
	char	hostname[32];
	char	gotname = FALSE;
	char	*cp;
	char	*queue;
	char	*user;
	int	fd;
	int	i;
	FILE	*tmpfp;
	struct stat sb;
	struct passwd *pw;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGALRM, dorestart );
	signal( SIGTERM, doflush );

	time (&timenow);
	ourpid = getpid();
	
	if (argc < 2) {
		retmsg (RP_FATAL, "uucpsend: incorrect arg count");
		err ("MDQS UUCPSEND: Arg count wrong (argc < 2)");
		cleanup();
		exit (11);
	}

	if(!gethostname(hostname, sizeof(hostname))) {	/* current hostname */
		gotname = TRUE;
	};

	for (i = 2; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
			case 'o':
				oldformat++;
				break;

			case 'h':	/* Name to send as requesting host */
				if (argv[i+1] == 0) {
					err (missingfmt, "host");
					exit (99);
				} else {
					i++;
					gotname = TRUE;
				}

				strncpy(hostname, argv[i], sizeof(hostname));
				break;

			default:
				retmsg (RP_FATAL, "Bad switch arg to uucpsend");
			}
		} else
			queue = argv[i];
	}

	sprintf (command, "uux -z - %s", argv[1]);

	if ((netfp = popen (command, "w")) == NULL) {
		retmsg (RP_RESTART, "Unable to popen() uux");
		exit (13);
	}
	mktemp (reqtmp);
	if ((fd = creat(reqtmp, 0600)) < 0) {
		err ("MDQS UUCPSEND: %s: %s", reqtmp, sys_errlist[errno]);
		cleanup();
		retmsg (RP_FATAL, "Cannot creat tempfile '%s'", reqtmp);
		exit (12);
	}
	close (fd);
	if ((tmpfp = fopen (reqtmp, "r+")) == NULL) {
		err ("MDQS UUCPSEND: Can't fopen reqtmp");
		unlink (reqtmp);
		retmsg (RP_FATAL, "Cannot open tempfile '%s'", reqtmp);
		exit (12);
	}
	unlink (reqtmp);
	
	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err ("MDQS UUCPSEND: Error reading request");
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}
	puthdr (&r, tmpfp);
	if ((pw = getpwuid(getuid())) == NULL)
		user = "UnknownUser";
	else
		user = pw->pw_name;

	if(gotname)
	   fprintf (netfp, "Site: %s\n", hostname);
	else
	   fprintf (netfp, "Site: %s\n", "UnknownHost");
if (!oldformat) {
	fprintf (netfp, "User: %s\n", user);
	fprintf (netfp, "Queue: %.*s\n", QUESIZ, r.r_que);
}
	fprintf (netfp, "Request: %.*s\n", QDIRSIZ, r.r_name);
	
	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'B':
			strcpy (newline, newtmp());
			fprintf (netfp, "File: %s\n", newline);
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			dumpfile (datafile);
			fprintf (tmpfp, "B%s\n", newline);
			break;

		case 'D':		/* Do NOT copy this line */
			if (workdir)
				free(workdir);
			workdir = strdup(&linebuf[1]);
			break;

		case 'F':
			strcpy (newline, newtmp());
			fprintf (netfp, "File: %s\n", newline);
			if( linebuf[1] == '/' ) {
				/* Absolute path */
				dumpfile (&linebuf[1]);
			} else {
				sprintf (datafile, "%s/%s", Qdatadir,
							&linebuf[1]);
				dumpfile (datafile);
			}
			fprintf (tmpfp, "I%s\n", newline);
			break;

		case 'H':
			strcpy (newline, newtmp());
			fprintf (netfp, "File: %s\n", newline);
			if( linebuf[1] == '/' ) {
				/* Absolute path */
				dumpfile (&linebuf[1]);
			} else {
				sprintf (datafile, "%s/%s", Qdatadir,
							&linebuf[1]);
				dumpfile (datafile);
			}
			fprintf (tmpfp, "B%s\n", newline);
			break;

		case 'I':
			strcpy (newline, newtmp());
			fprintf (netfp, "File: %s\n", newline);
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			dumpfile (datafile);
			fprintf (tmpfp, "I%s\n", newline);
			break;

		case 'O':
			options = atoi (&linebuf[1]);
			break;

		default:
			fputs (linebuf, tmpfp);
			putc ('\n', tmpfp);
			break;
		}
	}

	fflush (tmpfp);
	fseek (tmpfp, 0L, 0);
	fstat (fileno(tmpfp), &sb);
	fprintf (netfp, "File: %s\nSize: %ld\n", cp = newtmp(), sb.st_size);
	if (copycnt (tmpfp, netfp, cp, sb.st_size) != 0) {
		retmsg (RP_RESTART, "Error sending file to uux");
		exit (0);
	}
	fclose (tmpfp);
	fputs ("Done\n", netfp);
	fflush (netfp);
	if (ferror(netfp) || pclose (netfp) != 0)
		retmsg (RP_RESTART, "Error sending to uux");
	else
		retmsg (RP_OK, 0);
	exit (0);
}

/*
 *		D U M P F I L E ( )
 *  Dump a file onto the network with headers and trailers
 */
dumpfile( file )
char *file;
{
	FILE	*df;
	struct stat sb;

	if ((df = fopen (file, "r")) == NULL) {
		err ("MDQS UUCPSEND: Can't open %s", file);
		cleanup();
		retmsg (RP_FATAL, "Cannot open data file '%s'", file);
		exit (14);
	}
	fstat (fileno(df), &sb);
	fprintf (netfp, "Size: %ld\n", sb.st_size);
	if (copycnt (df, netfp, file, sb.st_size) != 0) {
		err ("MDQS UUCPSEND: Premature EOF on %s (%s)",
			file, sys_errlist[errno]);
		cleanup();
		retmsg (RP_FATAL, "Premature EOF on '%s'", file);
		exit (15);
	}
	fclose (df);
}

char *
newtmp ()
{
	static char	file[16];

	sprintf (file, "%lx%x", timenow++, ourpid, getuid());
	return (file);
}

cleanup ()
{
	char	datafile[PATHSIZ];
	char	linebuf[LINESIZ];
	struct	request r;
	register char 	*cp;

	/* Remove data files */
	rewind (stdin);
	gethdr (stdin, &r);
	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'B':
		case 'I':
			sprintf (datafile, "%s%s", Qdatadir, &linebuf[1]);
			unlink (datafile);
			break;
		case 'F':
			if (options & NOKEEP)
				unlink (&linebuf[1]);
		}
	}
}
