/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *			   Verset Emulator Queuer
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	20 Dec 82  DPK	Elinimated the fatal() function.
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>
#include "queue.h"

extern	int	Pri_min, Pri_max;
extern	char	*Tmpqdir;

/*
 *	Configurations Values for VERSET
 */
int	Verspri		= 64;		/* Default priority (in the middle) */
char	*Versforms	= "roll";	/* Dummy */
char	*Versqueue	= "verset";	/* Default queue for batch jobs */
char	*Sysmgr		= "mdqs";	/* Default orphanage for messages */
char	*Hostname	= "localhost";	/* Default local host name */

extern struct passwd *getpwuid();
extern char	*index();
extern char	*ckforms();
extern char	**findque();
extern char	*strdup();
extern FILE	*popen();
extern time_t	getdate();
extern int	cleanup();

char	*forms		= NULL;
char	*queue		= NULL;
char	*output		= NULL;
char	*username	= NULL;		/* Internet Return Address */
char	missingfmt[]	= "Missing %s argument";
char	dupfmt[]	= "Duplicate %s argument given\n";

int	priority	= 0;
int	qoptions	= 0;
int	error		= 0;
time_t	atime		= 0L;
time_t	timenow		= 0L;
struct	passwd *pw;
struct	request r;
FILE	*mkcontrol();
FILE	*mkdata();
FILE	*cf;			/* The control file */

struct paramtab paramtab[] = {
	"sysmgr",	STRPVAR,	&Sysmgr,		NULL,
	"hostname",	STRPVAR,	&Hostname,		NULL,
	"vers-forms",	STRPVAR,	&Versforms,		NULL,
	"vers-queue",	STRPVAR,	&Versqueue,		NULL,
	"vers-prior",	INTVAR,		(char **) &Verspri,	NULL,
	0,		0,		0,			0
};

/*
 *		M A I N ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	FILE	*df;			/* Data file, vers0 */
	char	dataname[QDIRSIZ];
	char	*arg;
	char	*cp;
	char	**qp;
	int	i;
	long	total;

	/*  We only act with the privileges of the REAL id  */
	r.r_gid = getgid();
	r.r_uid = getuid();
	time(&timenow);

	signal( SIGHUP, cleanup );
	signal( SIGQUIT, cleanup );
	signal( SIGINT, cleanup );

	rdconf( paramtab );
	umask (077);

	r.r_copies = 1;
	r.r_pri = Verspri;
	r.r_tstart = 0L;
	cf = mkcontrol ("QvsXXXXXX", r.r_name);
	puthdr (&r, cf);

	/* Write out the one line INTERNET address for notification */
	if( username == 0 ) {
		if ((pw = getpwuid (r.r_uid)) == NULL)
			fprintf( cf, "UUser Unknown <%s>\n", Sysmgr );
		else
			fprintf( cf, "U%s@%s\n", pw->pw_name, Hostname );
	} else
		fprintf( cf, "U%s\n", username );

	/* Dump the current directory so file references will work */
	dumpdir( cf );

	for (i = 1; i < argc; i++) {
		if (*argv[i] != '-') {
			dumparg( argv[i] );
			continue;
		}
		arg = argv[i];
		for( arg++; *arg; arg++ ) {
			switch (*arg) {
			case 'a':	/* Run after time */
				if (argv[i+1] == 0) {
					err (missingfmt, "time");
					error++;
					break;
				} else
					i++;
				if (atime != 0L) {
					err (dupfmt, "time");
					error++;
				} else if ((atime = getdate(argv[i], NULL)) == -1L) {
					err ("Bad time specification");
					error++;
				} else
					r.r_tstart = atime;
				break;

			case 'f':	/* What forms to print on */
				if (argv[i+1] == 0) {
					err (missingfmt, "forms");
					error++;
					break;
				} else
					i++;
				if (forms != 0) {
					err (dupfmt, "forms");
					error++;
				} else if ((forms = ckforms (argv[i])) == 0) {
					err ("Unknown form %s", argv[i]);
					error++;
				}
				break;

			case 'm':	/* Send mail upon completion */
				qoptions |= O_MAIL;
				break;

			case 'o':
				if (argv[i+1] == 0) {
					err (missingfmt, "output");
					error++;
					break;
				} else
					i++;
				if (output != 0) {
					err (dupfmt, "output");
					error++;
				}
				output = argv[i];
				break;

			case 'p':
				if (argv[i+1] == 0) {
					err (missingfmt, "priority");
					error++;
					break;
				} else
					i++;
				if (priority != 0) {
					err (dupfmt, "priority");
					error++;
				} else if (checknum(argv[i],
				    &priority, Pri_min, Pri_max) == -1) {
					fprintf (stderr,
				  "Priorities must be in the range %d to %d\n",
					  Pri_min, Pri_max);
					error++;
				} else
					r.r_pri = priority;
				break;

			case 'q':
				if (argv[i+1] == 0) {
					err (missingfmt, "queue");
					error++;
					break;
				} else
					i++;
				if (queue != 0) {
					err (dupfmt, "queue");
					error++;
				} else if((qp = findque( argv[i] )) == NULL) {
					err ("Unknown queue '%s'", argv[i]);
					error++;
				}
				queue = argv[i];
				break;

			case 'u':
				if (argv[i+1] == 0) {
					err (missingfmt, "user");
					error++;
					break;
				} else
					i++;
				if (username != 0) {
					err (dupfmt, "user");
					error++;
				}
				username = argv[i];
				break;

			case '?':
				error++;
				break;

			default:
				if (dumparg( argv[i] )) {
					unlink (r.r_name);
					exit(9);
				}
				goto donearg;
			}
		}
donearg:	continue;
	}
	
	if( error ) {
		fputs( "Usage:   verset [options] [troffoutput]\n", stderr );
		fputs( "Options: [-a time] [-g gfile] [-mn] [-o file]\n",  stderr );
		fputs( "         [-p pri] [-q queue] [-u useraddess]\n", stderr );
		exit (1);
	}

	if( queue == NULL )
		if((qp = findque( queue = Versqueue )) == NULL )
			bomb( "Default queue '%s' not configured\n", queue );

	while (*++qp)
		if(strncmp(*qp, "form=", 5) == SAME) {
			forms = strdup(*qp+5);
			break;
		}

	if( forms == NULL )
		forms = Versforms;

	if(( forms = ckforms( forms )) == NULL )
		bomb( "Forms '%s' not known\n", Versforms );

	r.r_options = qoptions;
	strncpy (r.r_que, queue, QUESIZ);
	strncpy (r.r_forms, forms, FORMSIZ);

	/* Write out the request structure */
	puthdr (&r, cf);
	fseek( cf, 0L, 2);			/* And, go back to the end */

	if( output )
		fprintf( cf, "P%s\n", output );

	/*
	 *  Read TROFF commands from the standard input.
	 */
	df = mkdata (timenow++, dataname);
	total = copy (stdin, df, "STDIN");
	fclose (df);
	fprintf (cf, "TSTDIN\n");
	fprintf (cf, "C%ld\n", total);
	fprintf (cf, "I%s\n", dataname);
	fclose (cf);
	if( ferror( cf ) || ferror( df ) ) {
		err( "Error creating queue files, aborting" );
		exit( 99 );
	}
	queueit(&r, 0);

	/**** Never Reached ****/
}

/*
 *		D U M P A R G ( )
 *  The string give is checked for newlines and then written into
 *  control file as an argument line ("A.....").
 */
dumparg( arg )
char	*arg;
{
	if( index( arg, '\n' )) {
		err("verset: argument '%s' contains a newline.");
		return (1);
	}
	fprintf( cf, "A%s\n", arg );
	return (0);
}

cleanup(signal)
int	signal;
{
	rmdata( cf );
	if( r.r_name[0] ) {
		char	longname[PATHSIZ];
	
		sprintf (longname, "%s/%s", Tmpqdir, r.r_name);
		unlink (longname);
	}
	_exit (signal);
}
